// objectdet.cpp : main function for object detection
//

#include "stdafx.h"

#include "common.h"
#include "adacascadeclass.h"
#include "data.h"
#include "detectobjects.h"
#include "dump.h"
#include "integralhistogram.h"
#include "prune_subwindows.h"
#include "profiling.h"
#include "localhistintegral.h"

#include <math.h>
#include <time.h>


int _tmain(int argc, _TCHAR* argv[])
{
	using namespace std;

    double dt;
    int64 t1,t2,t3,t01,t02;

	char imgfname[512]   = "images/000537.png";
	char classfname[512] = "classifiers/cabmodel_VOC06carside.xml";
	//char imgfname[512]   = "images/0000013650.jpg";
	//char classfname[512] = "classifiers/cabmodel_interm_nst40_RShast90train3_train_orienthistmod.xml";
    char resfname[512]   = "results/detrect.xml";
	uint nlevelsinoctave = 10;
	uint recomplevelstep = 1;
	float expvalinit     = 0.0;
	uint rotmode = 0;
	uint stretchmode = 0;
	uint flipmode = 0;
	uint robustmode = 0;
	int borderext = 0;
	int confthresh    = 5;
	uint min_stage_num = -1;
    bool streamingflag = false, saveresflag = false, quietflag = false, dispconfflag = false, pruneflag = true;

	//*** process command line arguments
	for (int i = 1; i < argc; i++){  /* Skip argv[0] (program name). */
		if (strcmp(argv[i], "-s") == 0){  // camera input
			streamingflag = true;
        } else
		if (strcmp(argv[i], "-i") == 0){  // image input
			if (i + 1 <= argc - 1){
                i++; strcpy(imgfname,argv[i]);
			}
        } else
		if (strcmp(argv[i], "-c") == 0){  // classifier file name
			if (i + 1 <= argc - 1){
                i++; strcpy(classfname,argv[i]);
            }
		} else
		if (strcmp(argv[i], "-o") == 0){  // file name for xml result output
			if (i + 1 <= argc - 1){
				saveresflag=true;
                i++; strcpy(resfname,argv[i]);
            }
		} else
		if (strcmp(argv[i], "-n") == 0){  // num. of pyramid levels in octave
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&nlevelsinoctave);
            }
		} else
		if (strcmp(argv[i], "-r") == 0){  // level-step for image processing
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&recomplevelstep);
            }
		} else
		if (strcmp(argv[i], "-e") == 0){  // exponent value for the 0 pyramid level
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%f",&expvalinit);
            }
		} else
		if (strcmp(argv[i], "-a") == 0){  // rotation mode
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&rotmode);
            }
		} else
		if (strcmp(argv[i], "-x") == 0){  // stretch mode
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&stretchmode);
            }
		} else
		if (strcmp(argv[i], "-f") == 0){  // flip mode
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&flipmode);
            }
		} else
		if (strcmp(argv[i], "-z") == 0){  // robust mode
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&robustmode);
            }
		} else
		if (strcmp(argv[i], "-b") == 0){  // flip mode
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&borderext);
            }
		} else
		if (strcmp(argv[i], "-t") == 0){  // threshold on detection confidence
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&confthresh);
            }
		} else
		if (strcmp(argv[i], "-d") == 0) { // display confidence flag
			dispconfflag = true;
		} else 
		if (strcmp(argv[i], "-p") == 0) { // no pruning flag -> detection confidence
										  // is the number of passed stages in this case
			pruneflag = false;
		} else 
		if (strcmp(argv[i], "-m") == 0){  // min. number of passed stages
			if (i + 1 <= argc - 1){
                i++; sscanf(argv[i],"%d",&min_stage_num);
            }
		} else
		if (strcmp(argv[i], "-q") == 0) { // quiet flag
			quietflag = true;
		} else {
			printf("Usage: objectdet\n");
			printf("         -s          : streaming input from camera\n");
			printf("         -i filename : image input from file\n");
			printf("         -c filename : classifier file name (xml)\n");
			printf("         -o filename : output xml file if output to disk is required\n");
			printf("         -n integer  : number of pyramid levels per octave (=10 default)\n");
			printf("         -r integer  : pyramid level step for image processing (=1 default)\n");
			printf("         -e double   : exponent value to define image scale at the finest\n");
			printf("                       pyramid level: scale_factor = 1/2^exp (=0 default)\n");
			printf("         -a integer  : rotation mode: 0 - rot. angle = 0 (default)\n");
			printf("                                      1 - rot. angle = [-pi/12, 0, pi/12]\n");
			printf("                                      2 - rot. angle = [-pi/6,  0, pi/6]\n");
			printf("                                      3 - rot. angle = [-pi/3, 0, pi/3]\n");
			printf("         -x integer  : stretch mode:  0 - x-stretch factor = 1 (default)\n");
			printf("                                      1 - x-stretch factor = [1, 1.5]\n");
			printf("                                      2 - x-stretch factor = [1, 2.0]\n");
			printf("                                      3 - x-stretch factor = [.5, 1, 2.0]\n");
			printf("         -f integer  : flip mode:     0 - no flip (default)\n");
			printf("                                      1 - no/x-axis flip\n");
			printf("                                      2 - no/y-axis flip\n");
			printf("                                      3 - no/x-/y-axis flip\n");
			printf("         -z integer  : robust mode:   0 - no hist. block stretching (default)\n");
			printf("                                      1 - hist. block stretching factor [1, 1.2]\n");
			printf("                                      2 - hist. block stretching factor [1, 1.2 1.4]\n");
			printf("         -b integer  : extension of an image in pixels (=0 default)\n");
			printf("         -t integer  : threshold on detection confidence (=5 default)\n");
			printf("         -d          : display detection confidence\n");
			printf("         -p          : no pruning (clustering) of detected windows\n");
			printf("         -m integer  : return windows which passed the min. number of stages\n");
			printf("                       (=-1: default -- passed all stages of the classifier)\n");
			printf("         -q          : quite flag disables graphical/textual output\n");
			exit(0);
		}
	}

	//*** initialize camera input
	CvCapture* capture = 0;
	IplImage *frame, *src = 0;
	int videodev=1;
	if (streamingflag){
		capture = cvCaptureFromCAM(videodev);	
		if(!capture){
			printf("cannot initialize video device '%d'",videodev);
			exit(0);
		}
		cvGrabFrame(capture);
		frame=cvRetrieveFrame(capture);
		if(!frame){
			printf("cannot read frame from video device '%d'",videodev);
			exit(0);
		}
		src = cvCreateImage( cvSize(frame->width,frame->height),IPL_DEPTH_8U, frame->nChannels );
		if( frame->origin == IPL_ORIGIN_TL )
			cvCopy( frame, src, 0 );
		else
			cvFlip( frame, src, 0 );
	}	
	//*** initialise image file input
	else {
		src=cvLoadImage(imgfname,0);
		if (!src) {
			printf("Could not read image %s\n",imgfname);
			exit(0);
		}
	}

	//*** load model
	Model model;
	xmlloadmodel(classfname,model);
    Tick("Model loading");
	//printf("Load %d-cascade classifier from %s\n",model.size(),classfname);

	//*** set image processing options
	PixClass pclass;
	CvMat *cmeans=NULL;
	CvMat *covmat=NULL;
	CvMat *cmeansnorm=NULL;
	if (!strcmp(model.begin()->histtype,"orienthistmod")){
		//*** use histograms of gradient orientation
		pclass = TEXTONS_SC12_GDORDER2;
	} else {
		//*** WARNING: no other histogram types are supported in this version
		printf("Histograms of type '%s' are not supported by the current version\n",model.begin()->histtype);
		exit(0);
		//*** read textons
		//char tfname[512] = "textons/textons_numberplate_set1_tnum12_scn2_ftrn5.xml";
		//CvFileStorage* tfsr=cvOpenFileStorage(tfname,NULL,CV_STORAGE_READ);
		//CvMat *cmeans=(CvMat*)cvRead(tfsr,cvGetFileNodeByName(tfsr,0,"cmeans"),NULL);
		//CvMat *covmat=(CvMat*)cvRead(tfsr,cvGetFileNodeByName(tfsr,0,"sqrtinvcovmat"),NULL);
		//CvMat *cmeansnorm=(CvMat*)cvRead(tfsr,cvGetFileNodeByName(tfsr,0,"cmeansnorm"),NULL);
		//cvReleaseFileStorage(&tfsr);
	}

	//*** set flip flags
	FlipFlagArray fliparray;
	fliparray.push_back(0);
	if ((flipmode==1) || (flipmode==3)) fliparray.push_back(1);
	if ((flipmode==2) || (flipmode==3)) fliparray.push_back(2);

	//*** set image scales 
	ScaleFctArray scarray;
	if (nlevelsinoctave<=0) nlevelsinoctave=5;
	if (recomplevelstep<1) recomplevelstep=1;
	for (double expval=expvalinit; expval<=10.0; expval+=(double)recomplevelstep*1.0/(double)nlevelsinoctave)
		scarray.push_back(1./pow(2.,expval));

	//*** set x/y ratio
	XYRatioArray xyratio;
	xyratio.push_back(1.0);
	if (stretchmode==1) xyratio.push_back(1.5);
	if (stretchmode==2) xyratio.push_back(2.0);
	if (stretchmode==3) { xyratio.push_back(0.5); xyratio.push_back(2.0); }

	//*** set rotation angles
    RotAngleArray anglearray;
	anglearray.push_back(0.0);
	double pi=3.14159265358979323846;
	if (stretchmode==1) { anglearray.push_back(pi/12); anglearray.push_back(-pi/12); }
	if (stretchmode==2) { anglearray.push_back(pi/6);  anglearray.push_back(-pi/6); }
	if (stretchmode==3) { anglearray.push_back(pi/3);  anglearray.push_back(-pi/3); }

	//*** set histogram blocks
	HistBlockSizeArray bsarray;
	double bszx=(double)(model.begin()->histblock_size_x);
	double bszy=(double)(model.begin()->histblock_size_y);
	for (int i=0; i<recomplevelstep; i++){
		double expval=(double)i/(double)nlevelsinoctave;
		double fct=pow(2.0,expval);
		bsarray.push_back(HistBlockSize(fct*bszx, fct*bszy));
		//*** set robust mode - stretching of hist. block size
		if (robustmode>0){
			bsarray.push_back(HistBlockSize(1.2*fct*bszx, fct*bszy));
			bsarray.push_back(HistBlockSize(fct*bszx, 1.2*fct*bszy));
		}
		if (robustmode>1){
			bsarray.push_back(HistBlockSize(1.4*fct*bszx, fct*bszy));
			bsarray.push_back(HistBlockSize(fct*bszx, 1.4*fct*bszy));
		}
	}
	

	if (!quietflag){
		printf("Starting object detection with parameters:\n");
		if (streamingflag) printf("  Camera iput\n");
		else               printf("  Image file iput:       %s\n",imgfname);
		                   printf("  Classifier file name:  %s\n",classfname);
		if (saveresflag)   printf("  Output in file:        %s\n",resfname);
		printf("  Number of pyramid levels in octave = %d\n",nlevelsinoctave);
		printf("  Pyramid level step for re-computing image descriptors = %d\n",recomplevelstep);
        printf("  Pre-scale original image factor = %1.1f\n",1./pow(2.,(double)expvalinit));
	}

	CvFont* font = new CvFont();
	char confstr[256], detstr[256];
	int conf; CvScalar col;
	cvInitFont(font,CV_FONT_HERSHEY_PLAIN,1.,1.);
	IplImage* imgdisp=cvCreateImage(cvSize(src->width,src->height),8,3);

	SubWindows subwindows;

	while (1){
		if (streamingflag){
			cvGrabFrame(capture);
			frame=cvRetrieveFrame(capture);
			if(!frame){
				printf("cannot read frame from video device '%d'",videodev);
				exit(0);
			}
			if( frame->origin == IPL_ORIGIN_TL ) cvCopy( frame, src, 0 );
			else cvFlip( frame, src, 0 );
		}

		//*** detect objects
		t1=cvGetTickCount();
		clock_t tc1 = clock ();
		subwindows = DetectObjects(src,model,cmeans,covmat,cmeansnorm,pclass,bsarray,scarray,fliparray,xyratio,anglearray,borderext,pruneflag,min_stage_num);
		t2=cvGetTickCount();
		clock_t tc2 = clock ();

		dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);
		//if (!quietflag) printf("detection time %1.3fsec.\n",dt);
		if (!quietflag) printf("detection time %1.3fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);


		//*** copy detection rectangles
		int rnum=(int)subwindows.size();
		CvMat* detrect;
		if (rnum>0) {
			detrect = cvCreateMat(rnum,5,CV_32FC1);
			int wcount=0;
			for (SubWindows::iterator wit = subwindows.begin(); wit != subwindows.end(); ++wit){
				//conf=wit->cluster_size;
				conf=wit->conf;
				cvSetReal2D(detrect, wcount, int(0), (double)wit->x);
				cvSetReal2D(detrect, wcount, int(1), (double)wit->y);
				cvSetReal2D(detrect, wcount, int(2), (double)wit->w);
				cvSetReal2D(detrect, wcount, int(3), (double)wit->h);
				cvSetReal2D(detrect, wcount, int(4), (double)conf);
				wcount++;
			}
		} else {
			detrect = cvCreateMat(1,5,CV_32FC1);
			cvSetReal2D(detrect, 0, int(4), (double)-1);
		}

		//*** display detection results
		if (!quietflag){
  		    // printf("%d SubWindows detected\n",rnum);
			if (src->nChannels==1) cvMerge(src,src,src,NULL,imgdisp);
			else cvCopy(src,imgdisp);
			if (rnum>0){
				for (int i=0; i<rnum; i++){
					double rx=cvGetReal2D(detrect, i, int(0));
					double ry=cvGetReal2D(detrect, i, int(1));
					double rw=cvGetReal2D(detrect, i, int(2));
					double rh=cvGetReal2D(detrect, i, int(3));
					double rc=cvGetReal2D(detrect, i, int(4));
					//if (rc<confthresh) col=cvScalar(100,100,0); else col=cvScalar(255,255,0);
					col=cvScalar(255,255,0);
					if (rc>=confthresh) {
						printf("%d, conf:%1.3f, rect:[%1.1f,%1.1f,%1.1f,%1.1f]\n",confthresh,rc,rx,ry,rw,rh);
						cvRectangle(imgdisp,cvPoint((int)rx,(int)ry),cvPoint((int)(rx+rw),(int)(ry+rh)),col,1);
						if (dispconfflag) sprintf(detstr,"%s|c:%d",model.begin()->label,(int)rc);
						else sprintf(detstr,"%s",model.begin()->label);
						cvPutText(imgdisp, detstr, cvPoint((int)rx,(int)ry),font,cvScalar(0,0,255));
						//cvPutText(imgdisp, model.begin()->label, cvPoint((int)rx,(int)ry),font,cvScalar(0,0,255));
						//sprintf(confstr,"conf: %d",(int)rc);
						//cvPutText(imgdisp, confstr, cvPoint((int)rx,(int)ry),font,cvScalar(0,0,255));
					}
				}
			}
			cvNamedWindow("Results",1);
			cvShowImage("Results",imgdisp);

		}

		if (saveresflag){
			if (!quietflag) printf("Saving detection rectangles to %s\n",resfname);
			CvFileStorage* fsw = cvOpenFileStorage(resfname,0,CV_STORAGE_WRITE);
			if (!fsw) {
				printf("Could not write results to %s\n",resfname);
				exit(0);
			}
			cvWrite(fsw,"detrect",detrect,cvAttrList(0,0));
			cvReleaseFileStorage( &fsw );  
		}
		cvReleaseMat(&detrect);

		if (cvWaitKey(10)>0) break;
		
		if (!streamingflag){
			if (!quietflag)
				cvWaitKey(0);
			break;
		}
	}

  cvReleaseMat(&cmeans);
  cvReleaseMat(&covmat);
  cvReleaseMat(&cmeansnorm);
  return 0;
}

